>From 1d69f62c5ab8b02d24ae924275f3f7ce343bb6ff Mon Sep 17 00:00:00 2001
From: Joseph Tulou <brezhoneg1@yahoo.fr>
Date: Mon, 26 Jan 2009 22:09:33 +0100
Subject: [PATCH] proof-of-concept vout reuse with restrictions

---
 src/video_output/video_output.c |  274 +++++++++++++++++++++++----------------
 1 files changed, 164 insertions(+), 110 deletions(-)

diff --git a/src/video_output/video_output.c b/src/video_output/video_output.c
index 27de388..9f32e21 100644
--- a/src/video_output/video_output.c
+++ b/src/video_output/video_output.c
@@ -77,6 +77,7 @@ static void VideoFormatImportRgb( video_format_t *, const picture_heap_t * );
 static void PictureHeapFixRgb( picture_heap_t * );
 
 static void     vout_Destructor   ( vlc_object_t * p_this );
+static int vout_Reset( vout_thread_t* p_vout, video_format_t *p_fmt );
 
 /* Object variables callbacks */
 static int DeinterlaceCallback( vlc_object_t *, char const *,
@@ -92,6 +93,8 @@ int vout_Snapshot( vout_thread_t *, picture_t * );
 /* Display media title in OSD */
 static void DisplayTitleOnOSD( vout_thread_t *p_vout );
 
+static void InitWindowSize( vout_thread_t *, unsigned *, unsigned * );
+
 /* Time during which the thread will sleep if it has nothing to
  * display (in micro-seconds) */
 #define VOUT_IDLE_SLEEP                 ((int)(0.020*CLOCK_FREQ))
@@ -155,127 +158,23 @@ vout_thread_t *__vout_Request( vlc_object_t *p_this, vout_thread_t *p_vout,
 {
     if( !p_fmt )
     {
-        /* Video output is no longer used.
-         * TODO: support for reusing video outputs with proper _thread-safe_
-         * reference handling. */
         if( p_vout )
             vout_CloseAndRelease( p_vout );
         return NULL;
     }
 
-    /* If a video output was provided, lock it, otherwise look for one. */
     if( p_vout )
     {
-        vlc_object_hold( p_vout );
-    }
-
-    /* TODO: find a suitable unused video output */
-
-    /* If we now have a video output, check it has the right properties */
-    if( p_vout )
-    {
-        char *psz_filter_chain;
-        vlc_value_t val;
-
-        vlc_mutex_lock( &p_vout->change_lock );
-
-        /* We don't directly check for the "vout-filter" variable for obvious
-         * performance reasons. */
-        if( p_vout->p->b_filter_change )
-        {
-            var_Get( p_vout, "vout-filter", &val );
-            psz_filter_chain = val.psz_string;
-
-            if( psz_filter_chain && !*psz_filter_chain )
-            {
-                free( psz_filter_chain );
-                psz_filter_chain = NULL;
-            }
-            if( p_vout->p->psz_filter_chain && !*p_vout->p->psz_filter_chain )
-            {
-                free( p_vout->p->psz_filter_chain );
-                p_vout->p->psz_filter_chain = NULL;
-            }
-
-            if( !psz_filter_chain && !p_vout->p->psz_filter_chain )
-            {
-                p_vout->p->b_filter_change = false;
-            }
-
-            free( psz_filter_chain );
-        }
-
-        if( p_vout->fmt_render.i_chroma != p_fmt->i_chroma ||
-            p_vout->fmt_render.i_width != p_fmt->i_width ||
-            p_vout->fmt_render.i_height != p_fmt->i_height ||
-            p_vout->p->b_filter_change )
-        {
-            vlc_mutex_unlock( &p_vout->change_lock );
-
-            /* We are not interested in this format, close this vout */
-            vout_CloseAndRelease( p_vout );
-            vlc_object_release( p_vout );
-            p_vout = NULL;
-        }
-        else
-        {
-            /* This video output is cool! Hijack it. */
-            if( p_vout->fmt_render.i_aspect != p_fmt->i_aspect )
-            {
-                /* Correct aspect ratio on change
-                 * FIXME factorize this code with other aspect ration related code */
-                unsigned int i_sar_num;
-                unsigned int i_sar_den;
-                unsigned int i_aspect;
-
-                i_aspect = p_fmt->i_aspect;
-                vlc_ureduce( &i_sar_num, &i_sar_den,
-                             p_fmt->i_sar_num, p_fmt->i_sar_den, 50000 );
-#if 0
-                /* What's that, it does not seems to be used correcly everywhere
-                 * beside the previous p_vout->fmt_render.i_aspect != p_fmt->i_aspect
-                 * should be fixed to use it too then */
-                if( p_vout->i_par_num > 0 && p_vout->i_par_den > 0 )
-                {
-                    i_sar_num *= p_vout->i_par_den;
-                    i_sar_den *= p_vout->i_par_num;
-                    i_aspect = i_aspect * p_vout->i_par_den / p_vout->i_par_num;
-                }
-#endif
-
-                if( i_sar_num > 0 && i_sar_den > 0 && i_aspect > 0 )
-                {
-                    p_vout->fmt_in.i_sar_num = i_sar_num;
-                    p_vout->fmt_in.i_sar_den = i_sar_den;
-                    p_vout->fmt_in.i_aspect  = i_aspect;
+        msg_Dbg( p_this, "reusing provided vout" );
 
-                    p_vout->fmt_render.i_sar_num = i_sar_num;
-                    p_vout->fmt_render.i_sar_den = i_sar_den;
-                    p_vout->fmt_render.i_aspect  = i_aspect;
+        vout_Reset( p_vout, p_fmt );
 
-                    p_vout->render.i_aspect   = i_aspect;
+        spu_Attach( p_vout->p_spu, p_this, true );
 
-                    p_vout->i_changes |= VOUT_ASPECT_CHANGE;
-
-                }
-            }
-            vlc_mutex_unlock( &p_vout->change_lock );
-
-            vlc_object_release( p_vout );
-        }
-
-        if( p_vout )
-        {
-            msg_Dbg( p_this, "reusing provided vout" );
-
-            spu_Attach( p_vout->p_spu, p_this, true );
-
-            vlc_object_detach( p_vout );
-            vlc_object_attach( p_vout, p_this );
-        }
+        vlc_object_detach( p_vout );
+        vlc_object_attach( p_vout, p_this );
     }
-
-    if( !p_vout )
+    else
     {
         msg_Dbg( p_this, "no usable vout present, spawning one" );
 
@@ -1733,3 +1632,158 @@ static void DisplayTitleOnOSD( vout_thread_t *p_vout )
     p_vout->p->psz_title = NULL;
 }
 
+/*
+ * Reset a previous vout and take new input format into account
+ */
+int vout_Reset( vout_thread_t* p_vout, video_format_t *p_fmt )
+{
+    vlc_mutex_lock( &p_vout->change_lock );
+    
+    /* clean up data from previous use of vout */
+    CleanThread( p_vout );
+
+    /* set data from new input */
+    unsigned int i_width = p_fmt->i_width;
+    unsigned int i_height = p_fmt->i_height;
+    vlc_fourcc_t i_chroma = p_fmt->i_chroma;
+    unsigned int i_aspect = p_fmt->i_aspect;
+
+    if( i_width <= 0 || i_height <= 0 || i_aspect <= 0 )
+        return NULL;
+
+    vlc_ureduce( &p_fmt->i_sar_num, &p_fmt->i_sar_den,
+                 p_fmt->i_sar_num, p_fmt->i_sar_den, 50000 );
+    if( p_fmt->i_sar_num <= 0 || p_fmt->i_sar_den <= 0 )
+        return NULL;
+
+    /* Initialize pictures - translation tables and functions
+     * will be initialized later in InitThread */
+    for( int i_index = 0; i_index < 2 * VOUT_MAX_PICTURES + 1; i_index++)
+    {
+        p_vout->p_picture[i_index].i_status = FREE_PICTURE;
+        p_vout->p_picture[i_index].i_type   = EMPTY_PICTURE;
+        p_vout->p_picture[i_index].b_slow   = 0;
+    }
+
+    /* No images in the heap */
+    p_vout->i_heap_size = 0;
+
+    /* Initialize the rendering heap */
+    I_RENDERPICTURES = 0;
+
+    p_vout->fmt_render        = *p_fmt;   /* FIXME palette */
+    p_vout->fmt_in            = *p_fmt;   /* FIXME palette */
+
+    p_vout->render.i_width    = i_width;
+    p_vout->render.i_height   = i_height;
+    p_vout->render.i_chroma   = i_chroma;
+    p_vout->render.i_aspect   = i_aspect;
+
+    p_vout->render.i_rmask    = p_fmt->i_rmask;
+    p_vout->render.i_gmask    = p_fmt->i_gmask;
+    p_vout->render.i_bmask    = p_fmt->i_bmask;
+
+    p_vout->render.i_last_used_pic = -1;
+    p_vout->render.b_allow_modify_pics = 1;
+
+    /* Zero the output heap */
+    I_OUTPUTPICTURES = 0;
+    p_vout->output.i_width    = 0;
+    p_vout->output.i_height   = 0;
+    // p_vout->output.i_chroma   = 0;
+    p_vout->output.i_aspect   = 0;
+
+    //p_vout->output.i_rmask    = 0;
+    //p_vout->output.i_gmask    = 0;
+    //p_vout->output.i_bmask    = 0;
+
+    /* Reset window size for new video */
+    //vout_IntfInit( p_vout );
+    InitWindowSize( p_vout, &p_vout->i_window_width,
+                    &p_vout->i_window_height );
+    vout_Control( p_vout, VOUT_SET_SIZE, p_vout->i_window_width,
+                  p_vout->i_window_height );
+
+    /* re-start vout */
+    InitThread( p_vout );
+
+    vlc_mutex_unlock( &p_vout->change_lock );
+
+    /**
+     * get a proper aspect-ratio for new input
+     * this call resets aspect-ratio to a predefined value
+     **/
+    var_SetString( p_vout, "aspect-ratio", "" );
+}
+
+
+
+/**
+ * FIXME : this is a copy of a local function from vout_intf.c 
+ * for the time being
+ **/
+void InitWindowSize( vout_thread_t *p_vout, unsigned *pi_width,
+                            unsigned *pi_height )
+{
+    vlc_value_t val;
+    int i_width, i_height;
+    uint64_t ll_zoom;
+
+#define FP_FACTOR 1000                             /* our fixed point factor */
+
+    var_Get( p_vout, "width", &val );
+    i_width = val.i_int;
+    var_Get( p_vout, "height", &val );
+    i_height = val.i_int;
+    var_Get( p_vout, "zoom", &val );
+    ll_zoom = (uint64_t)( FP_FACTOR * val.f_float );
+
+    if( i_width > 0 && i_height > 0)
+    {
+        *pi_width = (int)( i_width * ll_zoom / FP_FACTOR );
+        *pi_height = (int)( i_height * ll_zoom / FP_FACTOR );
+        goto initwsize_end;
+    }
+    else if( i_width > 0 )
+    {
+        *pi_width = (int)( i_width * ll_zoom / FP_FACTOR );
+        *pi_height = (int)( p_vout->fmt_in.i_visible_height * ll_zoom *
+            p_vout->fmt_in.i_sar_den * i_width / p_vout->fmt_in.i_sar_num /
+            FP_FACTOR / p_vout->fmt_in.i_visible_width );
+        goto initwsize_end;
+    }
+    else if( i_height > 0 )
+    {
+        *pi_height = (int)( i_height * ll_zoom / FP_FACTOR );
+        *pi_width = (int)( p_vout->fmt_in.i_visible_width * ll_zoom *
+            p_vout->fmt_in.i_sar_num * i_height / p_vout->fmt_in.i_sar_den /
+            FP_FACTOR / p_vout->fmt_in.i_visible_height );
+        goto initwsize_end;
+    }
+
+    if( p_vout->fmt_in.i_sar_num == 0 || p_vout->fmt_in.i_sar_den == 0 ) {
+        msg_Warn( p_vout, "fucked up aspect" );
+        *pi_width = (int)( p_vout->fmt_in.i_visible_width * ll_zoom / FP_FACTOR );
+        *pi_height = (int)( p_vout->fmt_in.i_visible_height * ll_zoom /FP_FACTOR);
+    }
+    else if( p_vout->fmt_in.i_sar_num >= p_vout->fmt_in.i_sar_den )
+    {
+        *pi_width = (int)( p_vout->fmt_in.i_visible_width * ll_zoom *
+            p_vout->fmt_in.i_sar_num / p_vout->fmt_in.i_sar_den / FP_FACTOR );
+        *pi_height = (int)( p_vout->fmt_in.i_visible_height * ll_zoom
+            / FP_FACTOR );
+    }
+    else
+    {
+        *pi_width = (int)( p_vout->fmt_in.i_visible_width * ll_zoom
+            / FP_FACTOR );
+        *pi_height = (int)( p_vout->fmt_in.i_visible_height * ll_zoom *
+            p_vout->fmt_in.i_sar_den / p_vout->fmt_in.i_sar_num / FP_FACTOR );
+    }
+
+initwsize_end:
+    msg_Dbg( p_vout, "window size: %dx%d", p_vout->i_window_width,
+             p_vout->i_window_height );
+
+#undef FP_FACTOR
+}
-- 
1.5.2.5

